@@ -1,4 +1,4 @@
-/* $PostgreSQL: pgsql/src/interfaces/ecpg/ecpglib/execute.c,v 1.75 2008/01/15 10:31:47 meskes Exp $ */
+/* $PostgreSQL: pgsql/src/interfaces/ecpg/ecpglib/execute.c,v 1.76.2.1 2008/03/01 03:26:44 tgl Exp $ */
 
 /*
  * The aim is to get a simpler inteface to the database routines.
@@ -122,13 +122,15 @@ next_insert(char *text, int pos, bool questionmarks)
 			string = string ? false : true;
 		else if (!string)
 		{
-			if (text[p] == '$' && isdigit(text[p + 1]))
+			if (text[p] == '$' && isdigit((unsigned char) text[p + 1]))
 			{
 				/* this can be either a dollar quote or a variable */
 				int			i;
 
-				for (i = p + 1; isdigit(text[i]); i++);
-				if (!isalpha(text[i]) && isascii(text[i]) && text[i] != '_')
+				for (i = p + 1; isdigit((unsigned char) text[i]); i++)
+					/* empty loop body */ ;
+				if (!isalpha((unsigned char) text[i]) &&
+					isascii((unsigned char) text[i]) && text[i] != '_')
 					/* not dollar delimited quote */
 					return p;
 			}
@@ -679,12 +681,12 @@ ecpg_store_input(const int lineno, const bool force_indicator, const struct vari
 					strcpy(mallocedval, "array [");
 
 					for (element = 0; element < asize; element++)
-						sprintf(mallocedval + strlen(mallocedval), "%.14g,", ((float *) var->value)[element]);
+						sprintf(mallocedval + strlen(mallocedval), "%.15g,", ((float *) var->value)[element]);
 
 					strcpy(mallocedval + strlen(mallocedval) - 1, "]");
 				}
 				else
-					sprintf(mallocedval, "%.14g", *((float *) var->value));
+					sprintf(mallocedval, "%.15g", *((float *) var->value));
 
 				*tobeinserted_p = mallocedval;
 				break;
@@ -698,12 +700,12 @@ ecpg_store_input(const int lineno, const bool force_indicator, const struct vari
 					strcpy(mallocedval, "array [");
 
 					for (element = 0; element < asize; element++)
-						sprintf(mallocedval + strlen(mallocedval), "%.14g,", ((double *) var->value)[element]);
+						sprintf(mallocedval + strlen(mallocedval), "%.15g,", ((double *) var->value)[element]);
 
 					strcpy(mallocedval + strlen(mallocedval) - 1, "]");
 				}
 				else
-					sprintf(mallocedval, "%.14g", *((double *) var->value));
+					sprintf(mallocedval, "%.15g", *((double *) var->value));
 
 				*tobeinserted_p = mallocedval;
 				break;
@@ -1392,7 +1394,7 @@ ecpg_execute(struct statement * stmt)
 					if (PQresultStatus(results) == PGRES_COMMAND_OK)
 						ecpg_log("ecpg_execute line %d: Got PGRES_COMMAND_OK after PGRES_COPY_OUT\n", stmt->lineno);
 					else
-						ecpg_log("ecpg_execute line %d: Got error after PGRES_COPY_OUT: %s", PQresultErrorMessage(results));
+						ecpg_log("ecpg_execute line %d: Got error after PGRES_COPY_OUT: %s", stmt->lineno, PQresultErrorMessage(results));
 				}
 				break;
 			}
@@ -1489,8 +1491,13 @@ ECPGdo(const int lineno, const int compat, const int force_indicator, const char
 	 */
 	if (statement_type == ECPGst_prepnormal)
 	{
-		if (!ecpg_auto_prepare(lineno, connection_name, questionmarks, &prepname, query))
+		if (!ecpg_auto_prepare(lineno, connection_name, compat, questionmarks, &prepname, query))
+		{
+			setlocale(LC_NUMERIC, oldlocale);
+			ecpg_free(oldlocale);
+			va_end(args);
 			return (false);
+		}
 
 		/*
 		 * statement is now prepared, so instead of the query we have to
